from django.urls import reverse
from rest_framework.test import APITestCase
from tests.testapp.models import Book, Course, Student, Phone


class ViewTests(APITestCase):
    def setUp(self):
        self.book1 = Book.objects.create(
            title="Advanced Data Structures", author="S.Mobit"
        )
        self.book2 = Book.objects.create(
            title="Basic Data Structures", author="S.Mobit"
        )

        self.course = Course.objects.create(name="Data Structures",
                                            code="CS210")

        self.course.books.set([self.book1, self.book2])

        self.student = Student.objects.create(name="Yezy",
                                              age=24,
                                              course=self.course)

        self.phone1 = Phone.objects.create(
            number="076711110", type="Office", student=self.student
        )
        self.phone2 = Phone.objects.create(
            number="073008880", type="Home", student=self.student
        )

    def tearDown(self):
        Book.objects.all().delete()
        Course.objects.all().delete()
        Student.objects.all().delete()

    # *************** retrieve tests **************

    def test_retrieve_with_flat_query(self):
        url = reverse("book-detail", args=[self.book1.id])
        response = self.client.get(url + "?query={title, author}",
                                   format="json")

        self.assertEqual(
            response.data,
            {
                "title": "Advanced Data Structures",
                "author": "S.Mobit"
            }
        )

    def test_retrieve_with_nested_flat_query(self):
        url = reverse("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={name, age, course{name}}", format="json"
        )

        self.assertEqual(
            response.data,
            {"name": "Yezy", "age": 24, "course": {"name": "Data Structures"}},
        )

    def test_retrieve_with_nested_iterable_query(self):
        url = reverse("course-detail", args=[self.course.id])
        response = self.client.get(
            url + "?query={name, code, books{title}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS210",
                "books": [
                    {"title": "Advanced Data Structures"},
                    {"title": "Basic Data Structures"},
                ],
            },
        )

    def test_retrieve_reverse_relation_with_nested_iterable_query(self):
        url = reverse("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={name, age, phone_numbers{number}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 24,
                "phone_numbers": [
                    {"number": "076711110"},
                    {"number": "073008880"}
                ],
            },
        )

    # *************** list tests **************

    def test_list_with_flat_query(self):
        url = reverse("book-list")
        response = self.client.get(url + "?query={title, author}",
                                   format="json")

        self.assertEqual(
            response.data,
            [
                {"title": "Advanced Data Structures", "author": "S.Mobit"},
                {"title": "Basic Data Structures", "author": "S.Mobit"},
            ],
        )

    def test_list_with_nested_flat_query(self):
        url = reverse("student-list")
        response = self.client.get(
            url + "?query={name, age, course{name}}", format="json"
        )

        self.assertEqual(
            response.data,
            [{
                "name": "Yezy",
                "age": 24,
                "course": {
                    "name": "Data Structures"
                }
            }],
        )

    def test_list_with_nested_iterable_query(self):
        url = reverse("course-list")
        response = self.client.get(
            url + "?query={name, code, books{title}}", format="json"
        )

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {"title": "Advanced Data Structures"},
                        {"title": "Basic Data Structures"},
                    ],
                }
            ],
        )

    def test_list_reverse_relation_with_nested_iterable_query(self):
        url = reverse("student-list")
        response = self.client.get(
            url + "?query={name, age, phone_numbers{number}}", format="json"
        )

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Yezy",
                    "age": 24,
                    "phone_numbers": [
                        {"number": "076711110"},
                        {"number": "073008880"}
                    ],
                }
            ],
        )



from django.db import models


class Book(models.Model):
    title = models.CharField(max_length=50)
    author = models.CharField(max_length=50)


class Course(models.Model):
    name = models.CharField(max_length=50)
    code = models.CharField(max_length=30)
    books = models.ManyToManyField(Book, blank=True, related_name="courses")


class Student(models.Model):
    name = models.CharField(max_length=50)
    age = models.IntegerField()
    course = models.ForeignKey(Course, on_delete=models.CASCADE, related_name="students")


class Phone(models.Model):
    number = models.CharField(max_length=15)
    type = models.CharField(max_length=50)
    student = models.ForeignKey(Student, on_delete=models.CASCADE, related_name="phone_numbers")



def get_flattened_subclasses(class_):
    for subclass in class_.__subclasses__():
        yield subclass
        inner_subclasses = get_flattened_subclasses(subclass)
        for inner_subclass in inner_subclasses:
            yield inner_subclass


def get_model_serializer_subclasses(class_):
    subclasses = get_flattened_subclasses(class_)
    for subclass in subclasses:
        try:
            subclass.Meta.model
            yield subclass
        except AttributeError:
            pass



from rest_framework import viewsets
from tests.testapp.models import Book, Course, Student
from tests.testapp.serializers import (BookSerializer,
                                       CourseSerializer,
                                       StudentSerializer)

from drf_lighten.views import DynamicFieldsMixin


class BookViewSet(DynamicFieldsMixin, viewsets.ModelViewSet):
    serializer_class = BookSerializer
    queryset = Book.objects.all()


class CourseViewSet(DynamicFieldsMixin, viewsets.ModelViewSet):
    serializer_class = CourseSerializer
    queryset = Course.objects.all()


class StudentViewSet(DynamicFieldsMixin, viewsets.ModelViewSet):
    serializer_class = StudentSerializer
    queryset = Student.objects.all()



import operator

from django.conf import settings
from rest_framework import relations, serializers

from .parsers import split
from .utils import get_model_serializer_subclasses

try:
    string_type = basestring
except NameError:
    string_type = str


class DynamicFieldsMixin(object):
    def __init__(self, *args, **kwargs):
        fields = kwargs.pop("fields", None)
        exclude = kwargs.pop("exclude", None)

        super(DynamicFieldsMixin, self).__init__(*args, **kwargs)

        if fields is not None:
            self.lighten(fields, "fields")
        elif exclude is not None:
            self.lighten(exclude, "exclude")

    @classmethod
    def get_model_serializer(cls, model):
        subclasses = get_model_serializer_subclasses(cls)
        for subclass in subclasses:
            try:
                if subclass.Meta.model == model:
                    return subclass
            except AttributeError:
                pass

    def lighten(self, entries, argument):
        if argument == "fields":
            field_selector = operator.sub
        else:
            field_selector = operator.and_

        field_strings, field_dictionary = self.split_fields(entries)

        for field_name, field_entry in field_dictionary.items():
            if field_name in self.fields:
                self.pass_down_structure(field_name, field_entry, argument)

        subset = set(field_strings)
        never_pop = set(field_dictionary.keys())
        existing = set(self.fields.keys())

        for field_name in field_selector(existing, subset) - never_pop:
            self.fields.pop(field_name, None)

    def split_fields(self, fields):
        strings, dictionary = split(fields)
        if strings == ["*"]:
            strings = set(self.fields.keys()) - set(dictionary.keys())
        return strings, dictionary

    def get_kwargs(self, field):
        field_name = field.field_name
        inherit = (
            set(relations.MANY_RELATION_KWARGS)
            .union(set(serializers.LIST_SERIALIZER_KWARGS))
            .difference({"html_cutoff_text", "html_cutoff"})
        )

        kwargs = {}
        for attribute_name in inherit:
            try:
                kwargs[attribute_name] = getattr(field, attribute_name)
            except AttributeError:
                pass

        if kwargs["source"] in ["", field_name]:
            kwargs.pop("source")

        kwargs["context"] = kwargs.get("context") or self.context
        return kwargs

    def get_field_and_kwargs(self, field_name):
        field = self.fields[field_name]
        many = True
        if isinstance(field, serializers.ListSerializer):
            field = field.child
        elif isinstance(field, serializers.ManyRelatedField):
            field.child_relation.source = field.source
            field.child_relation.source_attrs = field.source_attrs
            field.child_relation.field_name = field.field_name
            field = field.child_relation
        else:
            many = False

        kwargs = self.get_kwargs(field)
        kwargs["many"] = many

        return field, kwargs

    def get_expanding_serializer(self, field, **kwargs):
        field_name = field.field_name
        expansion_dict_attribute = getattr(
            settings, "DRF_LIGHTEN_EXPANSION_CONFIG", "expandable_fields"
        )
        expansion_dict = getattr(self.Meta, expansion_dict_attribute, {})
        try:
            class_, exp_args, exp_kwargs = expansion_dict[field_name]
        except (ValueError, TypeError):
            class_, exp_args, exp_kwargs = expansion_dict[field_name], (), {}
        except KeyError:
            current_model = self.Meta.model
            model_field = current_model._meta.get_field(field.source)
            target_model = model_field.target_field.model
            try:
                class_ = DynamicFieldsMixin.get_model_serializer(target_model)
                if class_ is None:
                    raise ValueError
            except ValueError:
                try:
                    from drf_generator.serializers import serializer_factory
                except ImportError:
                    return field
                class_ = serializer_factory(target_model)
            exp_args, exp_kwargs = (), {}
        exp_kwargs.update(kwargs)
        return class_(*exp_args, **exp_kwargs)

    def pass_down_structure(self, field_name, field_entry, arg_name):
        field, kwargs = self.get_field_and_kwargs(field_name)
        kwargs[arg_name] = field_entry
        if not hasattr(field, "fields"):
            new_field = self.get_expanding_serializer(field, **kwargs)
        else:
            try:
                return field.lighten(field_entry, arg_name)
            except AttributeError:
                return

        if getattr(new_field, "source"):
            return
        self.fields[field_name] = new_field



import json
import re


def re_parser(string):
    step_0 = string.replace(' ', '').replace('{', '[').replace('}', ']')
    step_1 = re.sub(r'(\w+)\{', r'{"\1": [', step_0)
    step_2 = re.sub(r'(\w+)\[', r'{"\1": [', step_1)
    step_3 = re.sub(r'](?=.)', r']}', step_2)
    step_4 = re.sub(r'(?<=\[|,)(\*|\w+)(?=,|])', r'"\1"', step_3)
    return json.loads(step_4)


def recursive_parser(string):
    pass


def split(parsed):
    strings = []
    dictionary = {}

    for field_entry in parsed:
        if isinstance(field_entry, dict):
            dictionary.update(field_entry)
        else:
            strings.append(field_entry)

    return strings, dictionary


def preprocess(parsed):
    strings, dictionary = split(parsed)
    if dictionary:
        return strings + [dictionary]
    return strings


def validate_star(parsed):
    strings, dictionary = split(parsed)

    for inner_entry in dictionary.values():
        validate_star(inner_entry)

    if '*' in strings and len(strings) != 1:
        raise ValueError


def parse(string):
    try:
        parsed = json.loads(string)
    except (ValueError, TypeError):
        parsed = re_parser(string)

    parsed = preprocess(parsed)
    validate_star(parsed)
    return parsed



from rest_framework import serializers
from tests.testapp.models import Book, Course, Phone, Student

from drf_lighten.serializers import DynamicFieldsMixin


class BookSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    class Meta:
        model = Book
        fields = '__all__'


class CourseSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    books = BookSerializer(many=True, read_only=True)

    class Meta:
        model = Course
        fields = '__all__'


class PhoneSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    class Meta:
        model = Phone
        fields = '__all__'


class StudentSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    course = CourseSerializer(many=False, read_only=True)
    phone_numbers = PhoneSerializer(many=True, read_only=True)

    class Meta:
        model = Student
        fields = '__all__'
